@@ -7,9 +7,11 @@ import pathlib
 import re
 import shutil
 import subprocess
-from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
+import tempfile
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union
 
 import packaging
+import subprocess_tee
 
 from ansible_compat.config import (
     AnsibleConfig,
@@ -50,6 +52,7 @@ class Runtime:
         min_required_version: Optional[str] = None,
         require_module: bool = False,
         max_retries: int = 0,
+        environ: Optional[Dict[str, str]] = None,
     ) -> None:
         """Initialize Ansible runtime environment.
 
@@ -69,18 +72,18 @@ class Runtime:
                                 also perform Python imports from Ansible.
         :param max_retries: Number of times it should retry network operations.
                             Default is 0, no retries.
+        :param environ: Environment dictionary to use, if undefined
+                        ``os.environ`` will be copied and used.
         """
         self.project_dir = project_dir or os.getcwd()
         self.isolated = isolated
         self.max_retries = max_retries
+        self.environ = environ or os.environ.copy()
         if isolated:
             self.cache_dir = get_cache_dir(self.project_dir)
         self.config = AnsibleConfig()
 
-        if (
-            min_required_version is not None
-            and packaging.version.Version(min_required_version) > self.version
-        ):
+        if not self.version_in_range(lower=min_required_version):
             raise RuntimeError(
                 f"Found incompatible version of ansible runtime {self.version}, instead of {min_required_version} or newer."
             )
@@ -99,7 +102,7 @@ class Runtime:
             raise RuntimeError("Unable to find Ansible python module.")
 
         ansible_module_version = packaging.version.parse(
-            ansible_release_module.__version__  # type: ignore
+            ansible_release_module.__version__
         )
         if ansible_module_version != self.version:
             raise RuntimeError(
@@ -114,19 +117,32 @@ class Runtime:
             shutil.rmtree(self.cache_dir, ignore_errors=True)
 
     def exec(
-        self, args: Union[str, List[str]], retry: bool = False
+        self,
+        args: Union[str, List[str]],
+        retry: bool = False,
+        tee: bool = False,
+        env: Optional[Dict[str, str]] = None,
+        cwd: Optional[str] = None,
     ) -> CompletedProcess:
         """Execute a command inside an Ansible environment.
 
         :param retry: Retry network operations on failures.
+        :param tee: Also pass captured stdout/stderr to system while running.
         """
+        if tee:
+            run_func: Callable[..., CompletedProcess] = subprocess_tee.run
+        else:
+            run_func = subprocess.run
+
         for _ in range(self.max_retries + 1 if retry else 1):
-            result = subprocess.run(
+            result = run_func(
                 args,
                 universal_newlines=True,
                 check=False,
                 stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE,
+                env=env or self.environ,
+                cwd=cwd,
             )
             if result.returncode == 0:
                 break
@@ -156,8 +172,24 @@ class Runtime:
         msg = "Unable to find a working copy of ansible executable."
         raise MissingAnsibleError(msg, proc=proc)
 
+    def version_in_range(
+        self, lower: Optional[str] = None, upper: Optional[str] = None
+    ) -> bool:
+        """Check if Ansible version is inside a required range.
+
+        The lower limit is inclusive and the upper one exclusive.
+        """
+        if lower and self.version < packaging.version.Version(lower):
+            return False
+        if upper and self.version >= packaging.version.Version(upper):
+            return False
+        return True
+
     def install_collection(
-        self, collection: str, destination: Optional[Union[str, pathlib.Path]] = None
+        self,
+        collection: str,
+        destination: Optional[Union[str, pathlib.Path]] = None,
+        force: bool = False,
     ) -> None:
         """Install an Ansible collection.
 
@@ -172,7 +204,7 @@ class Runtime:
 
         # ansible-galaxy before 2.11 fails to upgrade collection unless --force
         # is present, newer versions do not need it
-        if self.version < packaging.version.parse("2.11"):
+        if force or self.version_in_range(upper="2.11"):
             cmd.append("--force")
 
         if destination:
@@ -189,6 +221,35 @@ class Runtime:
             _logger.error(msg)
             raise InvalidPrerequisiteError(msg)
 
+    def install_collection_from_disk(
+        self, path: str, destination: Optional[Union[str, pathlib.Path]] = None
+    ) -> None:
+        """Build and install collection from a given disk path."""
+        if not self.version_in_range(upper="2.11"):
+            self.install_collection(path, destination=destination, force=True)
+            return
+        # older versions of ansible able unable to install without building
+        with tempfile.TemporaryDirectory() as tmp_dir:
+            cmd = [
+                "ansible-galaxy",
+                "collection",
+                "build",
+                "--output-path",
+                tmp_dir,
+                path,
+            ]
+            _logger.info("Running %s", " ".join(cmd))
+            run = self.exec(cmd, retry=False)
+            if run.returncode != 0:
+                _logger.error(run.stdout)
+                raise AnsibleCommandError(run)
+            for archive_file in os.listdir(tmp_dir):
+                self.install_collection(
+                    os.path.join(tmp_dir, archive_file),
+                    destination=destination,
+                    force=True,
+                )
+
     def install_requirements(self, requirement: str, retry: bool = False) -> None:
         """Install dependencies from a requirements.yml."""
         if not os.path.exists(requirement):
@@ -199,7 +260,7 @@ class Runtime:
                 f"{requirement} file is not a valid Ansible requirements file."
             )
 
-        if isinstance(reqs_yaml, list) or 'roles' in reqs_yaml:
+        if isinstance(reqs_yaml, list) or "roles" in reqs_yaml:
             cmd = [
                 "ansible-galaxy",
                 "role",
@@ -236,23 +297,43 @@ class Runtime:
                 raise AnsibleCommandError(run)
 
     def prepare_environment(
-        self, required_collections: Optional[Dict[str, str]] = None, retry: bool = False
+        self,
+        required_collections: Optional[Dict[str, str]] = None,
+        retry: bool = False,
+        install_local: bool = False,
+        offline: bool = False,
     ) -> None:
         """Make dependencies available if needed."""
         if required_collections is None:
             required_collections = {}
 
-        self.install_requirements("requirements.yml", retry=retry)
+        if not offline:
+            self.install_requirements("requirements.yml", retry=retry)
 
+        destination = f"{self.cache_dir}/collections" if self.cache_dir else None
         for name, min_version in required_collections.items():
             self.install_collection(
                 f"{name}:>={min_version}",
-                destination=f"{self.cache_dir}/collections" if self.cache_dir else None,
+                destination=destination,
             )
 
         self._prepare_ansible_paths()
-        # install role if current project looks like a standalone role
-        self._install_galaxy_role(self.project_dir, ignore_errors=True)
+
+        if not install_local:
+            return
+
+        if os.path.exists("galaxy.yml"):
+            # molecule scenario within a collection
+            self.install_collection_from_disk(".", destination=destination)
+        elif pathlib.Path().resolve().parent.name == "roles" and os.path.exists(
+            "../../galaxy.yml"
+        ):
+            # molecule scenario located within roles/<role-name>/molecule inside
+            # a collection
+            self.install_collection_from_disk("../..", destination=destination)
+        else:
+            # no collection, try to recognize and install a standalone role
+            self._install_galaxy_role(self.project_dir, ignore_errors=True)
 
     def require_collection(  # noqa: C901
         self,
@@ -266,10 +347,10 @@ class Runtime:
         collection before failing.
         """
         try:
-            ns, coll = name.split('.', 1)
+            ns, coll = name.split(".", 1)
         except ValueError as exc:
             raise InvalidPrerequisiteError(
-                "Invalid collection name supplied: %s" % name
+                f"Invalid collection name supplied: {name}%s"
             ) from exc
 
         paths: List[str] = self.config.collections_paths
@@ -286,19 +367,19 @@ class Runtime:
 
         for path in paths:
             collpath = os.path.expanduser(
-                os.path.join(path, 'ansible_collections', ns, coll)
+                os.path.join(path, "ansible_collections", ns, coll)
             )
             if os.path.exists(collpath):
-                mpath = os.path.join(collpath, 'MANIFEST.json')
+                mpath = os.path.join(collpath, "MANIFEST.json")
                 if not os.path.exists(mpath):
                     msg = f"Found collection at '{collpath}' but missing MANIFEST.json, cannot get info."
                     _logger.fatal(msg)
                     raise InvalidPrerequisiteError(msg)
 
-                with open(mpath, 'r') as f:
+                with open(mpath, "r", encoding="utf-8") as f:
                     manifest = json.loads(f.read())
                     found_version = packaging.version.parse(
-                        manifest['collection_info']['version']
+                        manifest["collection_info"]["version"]
                     )
                     if version and found_version < packaging.version.parse(version):
                         if install:
@@ -332,27 +413,30 @@ class Runtime:
             (roles_path, "roles", True),
         ]
 
-        if self.isolated:
-            alterations_list.extend(
-                [
-                    (roles_path, f"{self.cache_dir}/roles", False),
-                    (library_paths, f"{self.cache_dir}/modules", False),
-                    (collections_path, f"{self.cache_dir}/collections", False),
-                ]
-            )
+        alterations_list.extend(
+            [
+                (roles_path, f"{self.cache_dir}/roles", False),
+                (library_paths, f"{self.cache_dir}/modules", False),
+                (collections_path, f"{self.cache_dir}/collections", False),
+            ]
+            if self.isolated
+            else []
+        )
 
         for path_list, path, must_be_present in alterations_list:
-            if must_be_present and not os.path.exists(path):
-                continue
+            if not os.path.exists(path):
+                if must_be_present:
+                    continue
+                os.makedirs(path, exist_ok=True)
             if path not in path_list:
                 path_list.insert(0, path)
 
         if library_paths != self.config.DEFAULT_MODULE_PATH:
-            _update_env('ANSIBLE_LIBRARY', library_paths)
+            self._update_env("ANSIBLE_LIBRARY", library_paths)
         if collections_path != self.config.collections_paths:
-            _update_env(ansible_collections_path(), collections_path)
+            self._update_env(ansible_collections_path(), collections_path)
         if roles_path != self.config.default_roles_path:
-            _update_env('ANSIBLE_ROLES_PATH', roles_path)
+            self._update_env("ANSIBLE_ROLES_PATH", roles_path)
 
     def _install_galaxy_role(
         self, project_dir: str, role_name_check: int = 0, ignore_errors: bool = False
@@ -373,7 +457,7 @@ class Runtime:
         """
         yaml = None
         galaxy_info = {}
-        meta_filename = os.path.join(project_dir, 'meta', 'main.yml')
+        meta_filename = os.path.join(project_dir, "meta", "main.yml")
 
         if not os.path.exists(meta_filename):
             if ignore_errors:
@@ -381,8 +465,8 @@ class Runtime:
         else:
             yaml = yaml_from_file(meta_filename)
 
-        if yaml and 'galaxy_info' in yaml:
-            galaxy_info = yaml['galaxy_info']
+        if yaml and "galaxy_info" in yaml:
+            galaxy_info = yaml["galaxy_info"]
 
         fqrn = _get_role_fqrn(galaxy_info, project_dir)
 
@@ -396,7 +480,7 @@ class Runtime:
                     raise InvalidPrerequisiteError(msg)
         else:
             # when 'role-name' is in skip_list, we stick to plain role names
-            if 'role_name' in galaxy_info:
+            if "role_name" in galaxy_info:
                 role_namespace = _get_galaxy_role_ns(galaxy_info)
                 role_name = _get_galaxy_role_name(galaxy_info)
                 fqrn = f"{role_namespace}{role_name}"
@@ -410,7 +494,7 @@ class Runtime:
         target = pathlib.Path(project_dir).absolute()
         exists = link_path.exists() or link_path.is_symlink()
         if not exists or os.readlink(link_path) != str(target):
-            if exists:
+            if link_path.exists():
                 link_path.unlink()
             link_path.symlink_to(str(target), target_is_directory=True)
         _logger.info(
@@ -418,6 +502,21 @@ class Runtime:
             link_path,
         )
 
+    def _update_env(self, varname: str, value: List[str], default: str = "") -> None:
+        """Update colon based environment variable if needed.
+
+        New values are prepended to make sure they take precedence.
+        """
+        if not value:
+            return
+        orig_value = self.environ.get(varname, default)
+        if orig_value:
+            value = [*value, *orig_value.split(":")]
+        value_str = ":".join(value)
+        if value_str != self.environ.get(varname, ""):
+            self.environ[varname] = value_str
+            _logger.info("Set %s=%s", varname, value_str)
+
 
 def _get_role_fqrn(galaxy_infos: Dict[str, Any], project_dir: str) -> str:
     """Compute role fqrn."""
@@ -426,39 +525,20 @@ def _get_role_fqrn(galaxy_infos: Dict[str, Any], project_dir: str) -> str:
 
     if len(role_name) == 0:
         role_name = pathlib.Path(project_dir).absolute().name
-        role_name = re.sub(r'(ansible-|ansible-role-)', '', role_name).split(
+        role_name = re.sub(r"(ansible-|ansible-role-)", "", role_name).split(
             ".", maxsplit=2
         )[-1]
 
     return f"{role_namespace}{role_name}"
 
 
-def _update_env(varname: str, value: List[str], default: str = "") -> None:
-    """Update colon based environment variable if needed.
-
-    New values are added by inserting them to assure they take precedence
-    over the existing ones.
-    """
-    if value:
-        orig_value = os.environ.get(varname, default=default)
-        if orig_value:
-            # Prepend original or default variable content to custom content.
-            value = [*value, *orig_value.split(':')]
-        value_str = ":".join(value)
-        if value_str != os.environ.get(varname, ""):
-            os.environ[varname] = value_str
-            _logger.info("Added %s=%s", varname, value_str)
-
-
 def _get_galaxy_role_ns(galaxy_infos: Dict[str, Any]) -> str:
     """Compute role namespace from meta/main.yml, including trailing dot."""
-    role_namespace = galaxy_infos.get('namespace', "")
+    role_namespace = galaxy_infos.get("namespace", "")
     if len(role_namespace) == 0:
-        role_namespace = galaxy_infos.get('author', "")
+        role_namespace = galaxy_infos.get("author", "")
     if not isinstance(role_namespace, str):
-        raise AnsibleCompatError(
-            "Role namespace must be string, not %s" % role_namespace
-        )
+        raise AnsibleCompatError(f"Role namespace must be string, not {role_namespace}")
     # if there's a space in the name space, it's likely author name
     # and not the galaxy login, so act as if there was no namespace
     if not role_namespace or re.match(r"^\w+ \w+", role_namespace):
@@ -470,4 +550,4 @@ def _get_galaxy_role_ns(galaxy_infos: Dict[str, Any]) -> str:
 
 def _get_galaxy_role_name(galaxy_infos: Dict[str, Any]) -> str:
     """Compute role name from meta/main.yml."""
-    return galaxy_infos.get('role_name', "")
+    return galaxy_infos.get("role_name", "")
